<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width,initial-scale=1.0">
    <title>Meltano</title>
    <link rel="shortcut icon" href="/static/favicon.ico">
    
    <script>window.FLASK = JSON.parse('{{ jsContext | tojson | safe }}');</script>
    
  <link href="/static/css/app.a946c60b.css" rel="stylesheet"></head>
  <body>
    <noscript>
      <strong>We're sorry but webapp doesn't work properly without JavaScript enabled. Please enable it to continue.</strong>
    </noscript>
    <div id="app"></div>
    <!-- built files will be auto injected -->
  <script type="text/javascript" src="/static/js/chunk-vendors.3f88de42.js"></script><script type="text/javascript" src="/static/js/app.41a6442e.js"></script></body>
</html>
